@@ -14,7 +14,7 @@ import docker
 from .helpers import which
 from .server import DATA_FOLDER
 from .rpc import RpcError
-from .rpc_cache import BitcoinCLICached
+from .rpc import BitcoinCLI
 from .helpers import load_jsons
 
 logger = logging.getLogger(__name__)
@@ -42,7 +42,7 @@ class Btcd_conn:
     def get_cli(self):
         ''' returns a BitcoinCLI '''
         # def __init__(self, user, passwd, host="127.0.0.1", port=8332, protocol="http", path="", timeout=30, **kwargs):
-        cli = BitcoinCLICached(self.rpcuser, self.rpcpassword, host=self.ipaddress, port=self.rpcport)
+        cli = BitcoinCLI(self.rpcuser, self.rpcpassword, host=self.ipaddress, port=self.rpcport)
         cli.getblockchaininfo()
         return cli
 
@@ -66,6 +66,7 @@ class BitcoindController:
         if self.check_existing() != None:
             return self.check_existing()
 
+        logger.debug("Starting bitcoind")
         self._start_bitcoind(cleanup_at_exit)
 
         self.wait_for_bitcoind(self.rpcconn)
@@ -101,10 +102,12 @@ class BitcoindController:
         cli = self.get_cli()
         try:
             test3rdparty_cli = cli.wallet("test3rdparty")
+            test3rdparty_cli.getbalance()
         except RpcError as rpce:
             # return-codes:
             # https://github.com/bitcoin/bitcoin/blob/v0.15.0.1/src/rpc/protocol.h#L32L87
             if rpce.error_code == -18: # RPC_WALLET_NOT_FOUND
+                logger.debug("Creating test3rdparty wallet")
                 cli.createwallet("test3rdparty")
                 test3rdparty_cli = cli.wallet("test3rdparty")
             else:
@@ -162,7 +165,7 @@ class BitcoindController:
             if datadir == None:
                 datadir = tempfile.mkdtemp(prefix="bitcoind_datadir")
             btcd_cmd += " -datadir={} ".format(datadir)
-        logger.info(" constructed bitcoind-command: {}".format(btcd_cmd))
+        logger.debug("constructed bitcoind-command: %s",btcd_cmd)
         return btcd_cmd
 
 class BitcoindPlainController(BitcoindController):
@@ -187,6 +190,11 @@ class BitcoindPlainController(BitcoindController):
         if cleanup_at_exit:
             atexit.register(cleanup_bitcoind)
     
+    def stop_bitcoind(self):
+        # not necessary as the cleanup_bitcoind() will do it automatically!
+        # ToDo: Implement it nevertheless
+        pass
+    
     def check_existing(self):
         ''' other then in docker, we won't check on the "instance-level". This will return true if if a 
             bitcoind is running on the default port. 
