@@ -10,7 +10,7 @@ from .key import Key
 from functools import wraps
 from flask import g, request, redirect, url_for
 
-from flask import Flask, Blueprint, render_template, request, redirect, url_for, jsonify, flash
+from flask import Flask, Blueprint, render_template, request, redirect, url_for, jsonify, flash, send_file
 from flask_login import login_required, login_user, logout_user, current_user
 from flask_login.config import EXEMPT_METHODS
 
@@ -18,7 +18,8 @@ from flask_login.config import EXEMPT_METHODS
 from .devices.bitcoin_core import BitcoinCore
 from .helpers import (alias, get_devices_with_keys_by_type, hash_password, 
                       get_loglevel, get_version_info, run_shell, set_loglevel, 
-                      verify_password, bcur2base64, get_txid, generate_mnemonic)
+                      verify_password, bcur2base64, get_txid, generate_mnemonic,
+                      get_startblock_by_chain, fslock)
 from .specter import Specter
 from .specter_error import SpecterError
 from .wallet_manager import purposes
@@ -30,6 +31,8 @@ from io import BytesIO
 import traceback
 from .devices.electrum import b43_decode
 from binascii import b2a_base64
+from .tor_util import start_hidden_service, stop_hidden_services
+from stem.control import Controller
 
 from pathlib import Path
 env_path = Path('.') / '.flaskenv'
@@ -50,15 +53,42 @@ def server_error(e):
 @app.before_request
 def selfcheck():
     """check status before every request"""
+    if app.specter.cli is not None:
+        type(app.specter.cli).counter=0
     if app.config.get('LOGIN_DISABLED'):
         app.login('admin')
 
+
 ########## template injections #############
 @app.context_processor
 def inject_debug():
     ''' Can be used in all jinja2 templates '''
     return dict(debug=app.config['DEBUG'])
 
+
+@app.context_processor
+def inject_tor():
+    if app.config['DEBUG']:
+        return dict(tor_service_id='', tor_enabled=False)
+    if request.args.get('action', '') == 'stoptor' or request.args.get('action', '') == 'starttor':
+        if hasattr(current_user, 'is_admin') and current_user.is_admin:
+            try:
+                current_hidden_services = app.controller.list_ephemeral_hidden_services()
+            except Exception:
+                current_hidden_services = []
+            if request.args.get('action', '') == 'stoptor' and len(current_hidden_services) != 0:
+                stop_hidden_services(app)
+            if request.args.get('action', '') == 'starttor' and len(current_hidden_services) == 0:
+                try:
+                    start_hidden_service(app)
+                except Exception as e:
+                    flash('Failed to start Tor hidden service.\
+Make sure you have Tor running with ControlPort configured and try again.\
+Error returned: {}'.format(e), 'error')
+                    return dict(tor_service_id='', tor_enabled=False)
+    return dict(tor_service_id=app.tor_service_id, tor_enabled=app.tor_enabled)
+
+
 ################ routes ####################
 @app.route('/wallets/<wallet_alias>/combine/', methods=['GET', 'POST'])
 @login_required
@@ -155,9 +185,9 @@ def index():
         return redirect("/wallets/%s" % app.specter.wallet_manager.wallets[app.specter.wallet_manager.wallets_names[0]].alias)
 
     # TODO: add onboarding process
-    if len(app.specter.device_manager.devices) == 0:
-        # For now: can't do anything until a device is registered
-        return redirect("/new_device/")
+    # if len(app.specter.device_manager.devices) == 0:
+    #     # For now: can't do anything until a device is registered
+    #     return redirect("/new_device/")
 
     return render_template("base.jinja", specter=app.specter, rand=rand)
 
@@ -218,8 +248,12 @@ def register():
         password = hash_password(request.form['password'])
         otp = request.form['otp']
         user_id = alias(username)
-        if User.get_user(app.specter, user_id) or User.get_user_by_name(app.specter, username):
-            flash('Username is already taken, please choose another one', "error")
+        if User.get_user(app.specter, user_id) \
+                or User.get_user_by_name(app.specter, username):
+            flash(
+                'Username is already taken, please choose another one',
+                'error'
+            )
             return redirect('/register?otp={}'.format(otp))
         if app.specter.burn_new_user_otp(otp):
             config = {
@@ -233,18 +267,28 @@ def register():
             }
             user = User(user_id, username, password, config)
             user.save_info(app.specter)
+            flash(
+                'You have registered successfully, \
+please login with your new account to start using Specter'
+            )
             return redirect('/login')
         else:
-            flash('Invalid registration link, please request a new link from the node operator.', 'error')
+            flash(
+                'Invalid registration link, \
+please request a new link from the node operator.',
+                'error'
+            )
             return redirect('/register?otp={}'.format(otp))
     return render_template('register.jinja', specter=app.specter)
 
+
 @app.route('/logout', methods=['GET', 'POST'])
 def logout():
     logout_user()
-    flash('You were logged out',"info")
+    flash('You were logged out', "info")
     app.specter.clear_user_session()
-    return redirect("/login") 
+    return redirect("/login")
+
 
 @app.route('/settings/', methods=['GET'])
 @login_required
@@ -254,6 +298,7 @@ def settings():
     else:
         return redirect("/settings/general")
 
+
 @app.route('/settings/hwi', methods=['GET'])
 @login_required
 def hwi_settings():
@@ -266,6 +311,7 @@ def hwi_settings():
         rand=rand
     )
 
+
 @app.route('/settings/general', methods=['GET', 'POST'])
 @login_required
 def general_settings():
@@ -280,14 +326,101 @@ def general_settings():
         hwi_bridge_url = request.form['hwi_bridge_url']
         if current_user.is_admin:
             loglevel = request.form['loglevel']
-        
+
         if action == "save":
             if current_user.is_admin:
-                set_loglevel(app,loglevel)
+                set_loglevel(app, loglevel)
 
             app.specter.update_explorer(explorer, current_user)
             app.specter.update_hwi_bridge_url(hwi_bridge_url, current_user)
             app.specter.check()
+        elif action == "backup":
+            return send_file(
+                app.specter.specter_backup_file(),
+                attachment_filename='specter-backup.zip',
+                as_attachment=True
+            )
+        elif action == "restore":
+            restore_devices = json.loads(request.form['restoredevices'])
+            restore_wallets = json.loads(request.form['restorewallets'])
+            for device in restore_devices:
+                with fslock:
+                    with open(
+                        os.path.join(
+                            app.specter.device_manager.data_folder,
+                            "%s.json" % device['alias']
+                        ),
+                        "w"
+                    ) as file:
+                        file.write(json.dumps(device, indent=4))
+            app.specter.device_manager.update()
+
+            rescanning = False
+            for wallet in restore_wallets:
+                try:
+                    app.specter.wallet_manager.cli.createwallet(
+                        os.path.join(
+                            app.specter.wallet_manager.cli_path,
+                            wallet['alias']
+                        ),
+                        True
+                    )
+                except Exception as e:
+                    # if wallet already exists in Bitcoin Core
+                    # continue with the existing one
+                    if 'already exists' not in str(e):
+                        flash(
+                            'Failed to import wallet {}, error: {}'
+                            .format(wallet['name'], e),
+                            'error'
+                        )
+                        continue
+                with fslock:
+                    with open(
+                        os.path.join(
+                            app.specter.wallet_manager.working_folder,
+                            "%s.json" % wallet['alias']
+                        ),
+                        "w"
+                    ) as file:
+                        file.write(json.dumps(wallet, indent=4))
+                app.specter.wallet_manager.update()
+                try:
+                    wallet_obj = app.specter.wallet_manager.get_by_alias(
+                        wallet['alias']
+                    )
+                    try:
+                        wallet_obj.cli.rescanblockchain(
+                            wallet['blockheight']
+                            if 'blockheight' in wallet
+                            else get_startblock_by_chain(app.specter),
+                            timeout=1
+                        )
+                        app.logger.info("Rescanning Blockchain ...")
+                        rescanning = True
+                    except requests.exceptions.ReadTimeout:
+                        # this is normal behavior in our usecase
+                        pass
+                    except Exception as e:
+                        app.logger.error(
+                            "Exception while rescanning blockchain: {}".format(e)
+                        )
+                        flash(
+                            "Failed to perform rescan for wallet: {}".format(e),
+                            'error'
+                        )
+                    wallet_obj.getdata()
+                except Exception:
+                        flash(
+                            'Failed to import wallet {}'
+                            .format(wallet['name']),
+                            'error'
+                        )
+            flash('Specter data was successfully loaded from backup.', 'info')
+            if rescanning:
+                flash('Wallets are rescanning for transactions history.\n\
+This may take a few hours to complete.', 'info')
+
     return render_template(
         "settings/general_settings.jinja",
         explorer=explorer,
@@ -298,6 +431,7 @@ def general_settings():
         rand=rand
     )
 
+
 @app.route('/settings/bitcoin_core', methods=['GET', 'POST'])
 @login_required
 def bitcoin_core_settings():
@@ -312,17 +446,24 @@ def bitcoin_core_settings():
     port = rpc['port']
     host = rpc['host']
     protocol = 'http'
+    autodetect = rpc['autodetect']
+    datadir = rpc['datadir']
+    err = None
+
     if "protocol" in rpc:
         protocol = rpc["protocol"]
     test = None
     if request.method == 'POST':
         action = request.form['action']
         if current_user.is_admin:
+            autodetect = 'autodetect' in request.form
+            if autodetect:
+                datadir = request.form['datadir']
             user = request.form['username']
             passwd = request.form['password']
             port = request.form['port']
             host = request.form['host']
-            
+
         # protocol://host
         if "://" in host:
             arr = host.split("://")
@@ -330,14 +471,18 @@ def bitcoin_core_settings():
             host = arr[1]
 
         if action == "test":
-            test = app.specter.test_rpc(
-                user=user,
-                password=passwd,
-                port=port,
-                host=host,
-                protocol=protocol,
-                autodetect=False
-            )
+            try:
+                test = app.specter.test_rpc(
+                    user=user,
+                    password=passwd,
+                    port=port,
+                    host=host,
+                    protocol=protocol,
+                    autodetect=autodetect,
+                    datadir=datadir
+                )
+            except Exception as e:
+                err = 'Fail to connect to the node configured: {}'.format(e)
         elif action == "save":
             if current_user.is_admin:
                 app.specter.update_rpc(
@@ -346,13 +491,16 @@ def bitcoin_core_settings():
                     port=port,
                     host=host,
                     protocol=protocol,
-                    autodetect=False
+                    autodetect=autodetect,
+                    datadir=datadir
                 )
             app.specter.check()
 
     return render_template(
         "settings/bitcoin_core_settings.jinja",
         test=test,
+        autodetect=autodetect,
+        datadir=datadir,
         username=user,
         password=passwd,
         port=port,
@@ -360,9 +508,11 @@ def bitcoin_core_settings():
         protocol=protocol,
         specter=app.specter,
         current_version=current_version,
+        error=err,
         rand=rand
     )
 
+
 @app.route('/settings/auth', methods=['GET', 'POST'])
 @login_required
 def auth_settings():
@@ -372,7 +522,11 @@ def auth_settings():
     new_otp = -1
     users = None
     if current_user.is_admin and auth == "usernamepassword":
-        users = [user for user in User.get_all_users(app.specter) if not user.is_admin]
+        users = [
+            user
+            for user in User.get_all_users(app.specter)
+            if not user.is_admin
+        ]
     if request.method == 'POST':
         action = request.form['action']
 
@@ -388,7 +542,10 @@ def auth_settings():
             if specter_username:
                 if current_user.username != specter_username:
                     if User.get_user_by_name(app.specter, specter_username):
-                        flash('Username is already taken, please choose another one', "error")
+                        flash(
+                            'Username is already taken, please choose another one',
+                            "error"
+                        )
                         return render_template(
                             "settings/auth_settings.jinja",
                             auth=auth,
@@ -405,8 +562,17 @@ def auth_settings():
             if current_user.is_admin:
                 app.specter.update_auth(auth)
                 if auth == "rpcpasswordaspin" or auth == "usernamepassword":
+                    if auth == "usernamepassword":
+                        users = [
+                            user
+                            for user in User.get_all_users(app.specter)
+                            if not user.is_admin
+                        ]
+                    else:
+                        users = None
                     app.config['LOGIN_DISABLED'] = False
                 else:
+                    users = None
                     app.config['LOGIN_DISABLED'] = True
 
             app.specter.check()
@@ -668,16 +834,7 @@ def new_wallet(wallet_type):
             rescan_blockchain = 'rescanblockchain' in request.form
             if rescan_blockchain:
                 app.logger.info("Rescanning Blockchain ...")
-                if app.specter.info['chain'] == "main":
-                    if not app.specter.info['pruned'] or app.specter.info['pruneheight'] < 481824:
-                        startblock = 481824
-                    else:
-                        startblock = app.specter.info['pruneheight']
-                else:
-                    if not app.specter.info['pruned']:
-                        startblock = 0
-                    else:
-                        startblock = app.specter.info['pruneheight']
+                startblock = get_startblock_by_chain(app.specter)
                 try:
                     wallet.cli.rescanblockchain(startblock, timeout=1)
                 except requests.exceptions.ReadTimeout:
@@ -688,7 +845,7 @@ def new_wallet(wallet_type):
                     err = "%r" % e
                 wallet.getdata()
             return redirect("/wallets/%s/" % wallet.alias)
-    
+
     return render_template(
         "wallet/new_wallet/new_wallet.jinja",
         wallet_type=wallet_type,
@@ -700,6 +857,7 @@ def new_wallet(wallet_type):
         rand=rand
     )
 
+
 @app.route('/wallets/<wallet_alias>/')
 @login_required
 def wallet(wallet_alias):
@@ -714,20 +872,43 @@ def wallet(wallet_alias):
     else:
         return redirect("/wallets/%s/tx/" % wallet_alias)
 
+
+@app.route('/wallets_overview/')
+@login_required
+def wallets_overview():
+    app.specter.check()
+    idx = int(request.args.get('idx', default=0))
+    return render_template(
+        "wallet/wallets_overview.jinja",
+        idx=idx,
+        history=True,
+        specter=app.specter,
+        rand=rand
+    )
+
+
 @app.route('/wallets/<wallet_alias>/tx/')
 @login_required
 def wallet_tx(wallet_alias):
+    return redirect("/wallets/%s/tx/history" % wallet_alias)
+
+
+@app.route('/wallets/<wallet_alias>/tx/history/')
+@login_required
+def wallet_tx_history(wallet_alias):
     app.specter.check()
     try:
         wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)
     except SpecterError as se:
         app.logger.error("SpecterError while wallet_tx: %s" % se)
         return render_template("base.jinja", error=se, specter=app.specter, rand=rand)
-    return render_template("wallet/history/txs/wallet_tx.jinja", wallet_alias=wallet_alias, wallet=wallet, specter=app.specter, rand=rand)
+    idx = int(request.args.get('idx', default=0))
 
-@app.route('/wallets/<wallet_alias>/addresses/', methods=['GET', 'POST'])
+    return render_template("wallet/history/txs/wallet_tx.jinja", idx=idx, wallet_alias=wallet_alias, wallet=wallet, history=True, specter=app.specter, rand=rand)
+
+@app.route('/wallets/<wallet_alias>/tx/utxo/', methods=['GET', 'POST'])
 @login_required
-def wallet_addresses(wallet_alias):
+def wallet_tx_utxo(wallet_alias):
     app.specter.check()
     try:
         wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)
@@ -746,8 +927,7 @@ def wallet_addresses(wallet_alias):
                 for address in wallet.addresses_on_label(account):
                     wallet.setlabel(address, label)
                 wallet.getdata()
-    alladdresses = True if request.args.get('all') != 'False' else False
-    return render_template("wallet/history/addresses/wallet_addresses.jinja", wallet_alias=wallet_alias, wallet=wallet, alladdresses=alladdresses, viewtype=viewtype, specter=app.specter, rand=rand)
+    return render_template("wallet/history/utxo/wallet_utxo.jinja", wallet_alias=wallet_alias, wallet=wallet, history=False, viewtype=viewtype, specter=app.specter, rand=rand)
 
 @app.route('/wallets/<wallet_alias>/receive/', methods=['GET', 'POST'])
 @login_required
@@ -765,7 +945,7 @@ def wallet_receive(wallet_alias):
         elif action == "updatelabel":
             label = request.form['label']
             wallet.setlabel(wallet.address, label)
-    if wallet.tx_on_current_address > 0:
+    if wallet.is_current_address_used:
         wallet.getnewaddress()
     return render_template("wallet/receive/wallet_receive.jinja", wallet_alias=wallet_alias, wallet=wallet, specter=app.specter, rand=rand)
 
@@ -775,9 +955,23 @@ def fees(blocks):
     res = app.specter.estimatesmartfee(int(blocks))
     return res
 
-@app.route('/wallets/<wallet_alias>/send/new', methods=['GET', 'POST'])
+@app.route('/wallets/<wallet_alias>/send')
 @login_required
 def wallet_send(wallet_alias):
+    app.specter.check()
+    try:
+        wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)
+    except SpecterError as se:
+        app.logger.error("SpecterError while wallet_send: %s" % se)
+        return render_template("base.jinja", error=se, specter=app.specter, rand=rand)
+    if len(wallet.pending_psbts) > 0:
+        return redirect(url_for('wallet_sendpending', wallet_alias=wallet_alias))
+    else:
+        return redirect(url_for('wallet_sendnew', wallet_alias=wallet_alias))
+
+@app.route('/wallets/<wallet_alias>/send/new', methods=['GET', 'POST'])
+@login_required
+def wallet_sendnew(wallet_alias):
     app.specter.check()
     try:
         wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)
@@ -785,19 +979,25 @@ def wallet_send(wallet_alias):
         app.logger.error("SpecterError while wallet_send: %s" % se)
         return render_template("base.jinja", error=se, specter=app.specter, rand=rand)
     psbt = None
-    address = ""
-    label = ""
-    amount = 0
+    addresses = [""]
+    labels = [""]
+    amounts = [0]
     fee_rate = 0.0
     err = None
     if request.method == "POST":
         action = request.form['action']
         if action == "createpsbt":
-            address = request.form['address']
-            label = request.form['label']
-            if request.form['label'] != "":
-                wallet.setlabel(address, label)
-            amount = float(request.form['btc_amount'])
+            i = 0
+            addresses = []
+            labels = []
+            amounts = []
+            while 'address_{}'.format(i) in request.form:
+                addresses.append(request.form['address_{}'.format(i)])
+                amounts.append(float(request.form['btc_amount_{}'.format(i)]))
+                labels.append(request.form['label_{}'.format(i)])
+                if request.form['label_{}'.format(i)] != '':
+                    wallet.setlabel(addresses[i], labels[i])
+                i += 1
             subtract = bool(request.form.get("subtract", False))
             fee_unit = request.form.get('fee_unit')
             selected_coins = request.form.getlist('coinselect')
@@ -809,19 +1009,19 @@ def wallet_send(wallet_alias):
                     fee_rate = float(request.form.get('fee_rate'))
 
             try:
-                psbt = wallet.createpsbt(address, amount, subtract=subtract, fee_rate=fee_rate, fee_unit=fee_unit, selected_coins=selected_coins)
+                psbt = wallet.createpsbt(addresses, amounts, subtract=subtract, fee_rate=fee_rate, fee_unit=fee_unit, selected_coins=selected_coins)
                 if psbt is None:
                     err = "Probably you don't have enough funds, or something else..."
                 else:
                     # calculate new amount if we need to subtract
                     if subtract:
                         for v in psbt["tx"]["vout"]:
-                            if address in v["scriptPubKey"]["addresses"]:
-                                amount = v["value"]
+                            if addresses[0] in v["scriptPubKey"]["addresses"]:
+                                amounts[0] = v["value"]
             except Exception as e:
                 err = e
             if err is None:
-                return render_template("wallet/send/sign/wallet_send_sign_psbt.jinja", psbt=psbt, label=label, 
+                return render_template("wallet/send/sign/wallet_send_sign_psbt.jinja", psbt=psbt, labels=labels, 
                                                     wallet_alias=wallet_alias, wallet=wallet, 
                                                     specter=app.specter, rand=rand)
         elif action == "importpsbt":
@@ -831,12 +1031,12 @@ def wallet_send(wallet_alias):
             except Exception as e:
                 flash("Could not import PSBT: %s" % e, "error")
                 return redirect(url_for('wallet_importpsbt', wallet_alias=wallet_alias))
-            return render_template("wallet/send/sign/wallet_send_sign_psbt.jinja", psbt=psbt, label=label, 
+            return render_template("wallet/send/sign/wallet_send_sign_psbt.jinja", psbt=psbt, labels=labels, 
                                                 wallet_alias=wallet_alias, wallet=wallet, 
                                                 specter=app.specter, rand=rand)
         elif action == "openpsbt":
             psbt = ast.literal_eval(request.form["pending_psbt"])
-            return render_template("wallet/send/sign/wallet_send_sign_psbt.jinja", psbt=psbt, label=label, 
+            return render_template("wallet/send/sign/wallet_send_sign_psbt.jinja", psbt=psbt, labels=labels, 
                                                 wallet_alias=wallet_alias, wallet=wallet, 
                                                 specter=app.specter, rand=rand)
         elif action == 'deletepsbt':
@@ -868,10 +1068,10 @@ def wallet_send(wallet_alias):
             else:
                 signed_psbt = None
                 flash("Device already signed the PSBT", "error")
-            return render_template("wallet/send/sign/wallet_send_sign_psbt.jinja", signed_psbt=signed_psbt, psbt=psbt, label=label, 
+            return render_template("wallet/send/sign/wallet_send_sign_psbt.jinja", signed_psbt=signed_psbt, psbt=psbt, labels=labels, 
                                                 wallet_alias=wallet_alias, wallet=wallet, 
                                                 specter=app.specter, rand=rand)
-    return render_template("wallet/send/new/wallet_send.jinja", psbt=psbt, label=label, 
+    return render_template("wallet/send/new/wallet_send.jinja", psbt=psbt, labels=labels, 
                                                 wallet_alias=wallet_alias, wallet=wallet, 
                                                 specter=app.specter, rand=rand, error=err)
 
@@ -907,6 +1107,12 @@ def wallet_sendpending(wallet_alias):
                 app.logger.error("Could not delete Pending PSBT: %s" % e)
                 flash("Could not delete Pending PSBT!", "error")
     pending_psbts = wallet.pending_psbts
+    ######## Migration to multiple recipients format ###############
+    for psbt in pending_psbts:
+        if not isinstance(pending_psbts[psbt]['address'], list):
+            pending_psbts[psbt]['address'] = [pending_psbts[psbt]['address']]
+            pending_psbts[psbt]['amount'] = [pending_psbts[psbt]['amount']]
+    ###############################################################
     return render_template("wallet/send/pending/wallet_sendpending.jinja", pending_psbts=pending_psbts,
                                                 wallet_alias=wallet_alias, wallet=wallet, 
                                                 specter=app.specter) 
@@ -945,10 +1151,10 @@ def wallet_settings(wallet_alias):
             wallet.keypoolrefill(wallet.keypool, wallet.keypool + delta)
             wallet.keypoolrefill(wallet.change_keypool, wallet.change_keypool + delta, change=True)
             wallet.getdata()
-        elif action == "rebuildcache":
-            wallet.cli.cache.rebuild_cache()
         elif action == "deletewallet":
-            app.specter.wallet_manager.delete_wallet(wallet)
+            app.specter.wallet_manager.delete_wallet(
+                wallet, app.specter.bitcoin_datadir
+            )
             response = redirect(url_for('index'))
             return response
         elif action == "rename":
@@ -958,11 +1164,12 @@ def wallet_settings(wallet_alias):
             else:
                 app.specter.wallet_manager.rename_wallet(wallet, wallet_name)
 
-        return render_template("wallet/settings/wallet_settings.jinja", 
+        return render_template(
+            "wallet/settings/wallet_settings.jinja",
             wallet_alias=wallet_alias,
-            wallet=wallet, 
+            wallet=wallet,
             specter=app.specter,
-            rand=rand, 
+            rand=rand,
             error=error
         )
     else:
@@ -1045,7 +1252,11 @@ def device(device_alias):
             if len(wallets) != 0:
                 err = "Device could not be removed since it is used in wallets: {}.\nYou must delete those wallets before you can remove this device.".format([wallet.name for wallet in wallets])
             else:
-                app.specter.device_manager.remove_device(device, app.specter.wallet_manager)
+                app.specter.device_manager.remove_device(
+                    device,
+                    app.specter.wallet_manager,
+                    bitcoin_datadir=app.specter.bitcoin_datadir
+                )
                 return redirect("/")
         elif action == "delete_key":
             key = request.form['key']
@@ -1073,18 +1284,25 @@ def device(device_alias):
             device=device, device_alias=device_alias, purposes=purposes, wallets=wallets, error=err, specter=app.specter, rand=rand)
 
 
-
 ############### filters ##################
 
 @app.template_filter('datetime')
 def timedatetime(s):
     return format(datetime.fromtimestamp(s), "%d.%m.%Y %H:%M")
 
+
 @app.template_filter('btcamount')
 def btcamount(value):
     value = float(value)
     return "{:.8f}".format(value).rstrip("0").rstrip(".")
 
+
+@app.template_filter('bytessize')
+def bytessize(value):
+    value = float(value)
+    return '{:,.0f}'.format(value / float(1 << 30)) + " GB"
+
+
 def notify_upgrade():
     ''' If a new version is available, notifies the user via flash 
         that there is an upgrade to specter.desktop
