@@ -1,5 +1,9 @@
-import os, json, logging
-from .device import Device
+import os
+import json
+import logging
+import time
+import zipfile
+from io import BytesIO
 from .devices.coldcard import ColdCard
 from .devices.trezor import Trezor
 from .devices.ledger import Ledger
@@ -10,6 +14,7 @@ from .devices.generic import GenericDevice
 from .devices.electrum import Electrum
 from .devices.bitcoin_core import BitcoinCore
 from .helpers import alias, load_jsons, fslock
+from .rpc import get_default_datadir
 
 
 logger = logging.getLogger(__name__)
@@ -82,7 +87,7 @@ class DeviceManager:
             with open(fullpath, "w") as file:
                 file.write(json.dumps(device.json, indent=4))
 
-        self.update() # reload files
+        self.update()  # reload files
         return device
 
     def get_by_alias(self, device_alias):
@@ -91,8 +96,13 @@ class DeviceManager:
                 return self.devices[device_name]
         logger.error("Could not find Device %s" % device_alias)
 
-    def remove_device(self, device, wallet_manager=None):
+    def remove_device(
+        self,
+        device,
+        wallet_manager=None,
+        bitcoin_datadir=get_default_datadir()
+    ):
         os.remove(device.fullpath)
         if isinstance(device, BitcoinCore):
-            device.delete(wallet_manager)
+            device.delete(wallet_manager, bitcoin_datadir=bitcoin_datadir)
         self.update()
