@@ -6,6 +6,9 @@ from hwilib.serializations import PSBT, CTransaction
 from .bcur import bcur_decode
 import threading
 from io import BytesIO
+import re
+
+logger = logging.getLogger(__name__)
 
 # use this for all fs operations
 fslock = threading.Lock()
@@ -29,11 +32,12 @@ try:
 except:
     collectionsAbc = collections
 
-logger = logging.getLogger(__name__)
+
 def alias(name):
     name = name.replace(" ", "_")
     return "".join(x for x in name if x.isalnum() or x=="_").lower()
 
+
 def deep_update(d, u):
     for k, v in six.iteritems(u):
         dv = d.get(k, {})
@@ -45,6 +49,7 @@ def deep_update(d, u):
             d[k] = v
     return d
 
+
 def load_jsons(folder, key=None):
     files = [f for f in os.listdir(folder) if f.endswith(".json")]
     files.sort(key=lambda x: os.path.getmtime(os.path.join(folder, x)))
@@ -61,14 +66,18 @@ def load_jsons(folder, key=None):
             dd[d[key]] = d
     return dd
 
+
 BASE58_ALPHABET = b'123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'
 
+
 def double_sha256(s):
     return hashlib.sha256(hashlib.sha256(s).digest()).digest()
 
+
 def hash160(d):
     return hashlib.new('ripemd160', hashlib.sha256(d).digest()).digest()
 
+
 def encode_base58(s):
     # determine how many 0 bytes (b'\x00') s starts with
     count = 0
@@ -87,9 +96,11 @@ def encode_base58(s):
 
     return prefix + bytes(result)
 
+
 def encode_base58_checksum(s):
     return encode_base58(s + double_sha256(s)[:4]).decode('ascii')
 
+
 def decode_base58(s, num_bytes=82, strip_leading_zeros=False):
     num = 0
     for c in s.encode('ascii'):
@@ -105,11 +116,13 @@ def decode_base58(s, num_bytes=82, strip_leading_zeros=False):
             checksum, double_sha256(combined)[:4]))
     return combined[:-4]
 
+
 def convert_xpub_prefix(xpub, prefix_bytes):
     # Update xpub to specified prefix and re-encode
     b = decode_base58(xpub)
     return encode_base58_checksum(prefix_bytes + b[4:])
 
+
 def get_xpub_fingerprint(xpub):
     b = decode_base58(xpub)
     return hash160(b[-33:])[:4]
@@ -124,7 +137,7 @@ def which(program):
         it will simply return it if found and executable
         will raise an Exception if not found
     '''
-    
+
     def is_exe(fpath):
         return os.path.isfile(fpath) and os.access(fpath, os.X_OK)
 
@@ -133,22 +146,23 @@ def which(program):
         # https://pyinstaller.readthedocs.io/en/v3.3.1/runtime-information.html#using-sys-executable-and-sys-argv-0
         exec_location = os.path.join(sys._MEIPASS, program)
         if is_exe(exec_location):
-            logger.info("Found %s executable in %s" % (program, exec_location))
+            logger.debug("Found %s executable in %s" % (program, exec_location))
             return exec_location
 
     fpath, program_name = os.path.split(program)
     if fpath:
         if is_exe(program):
-            logger.info("Found %s executable in %s" % (program, program))
+            logger.debug("Found %s executable in %s" % (program, program))
             return program
     else:
         for path in os.environ["PATH"].split(os.pathsep):
             exe_file = os.path.join(path, program)
             if is_exe(exe_file):
-                logger.info("Found %s executable in %s" % (program, path))
+                logger.debug("Found %s executable in %s" % (program, path))
                 return exe_file
     raise Exception("Couldn't find executable %s" % program)
 
+
 # should work in all python versions
 def run_shell(cmd):
     """
@@ -166,7 +180,8 @@ def run_shell(cmd):
     except:
         return { "code": 0xf00dbabe, "out": b"", "err": b"Can't run subprocess" }
 
-def set_loglevel(app,loglevel_string):
+
+def set_loglevel(app, loglevel_string):
     logger.info("Setting Loglevel to %s" % loglevel_string)
     loglevels = {
         "WARN": logging.WARN,
@@ -176,8 +191,8 @@ def set_loglevel(app,loglevel_string):
     app.logger.setLevel(loglevels[loglevel_string])
     logger.setLevel(loglevels[loglevel_string])
 
+
 def get_loglevel(app):
-    
     loglevels = {
         logging.WARN : "WARN",
         logging.INFO : "INFO",
@@ -185,12 +200,16 @@ def get_loglevel(app):
     }
     return loglevels[app.logger.getEffectiveLevel()]
 
+
 def get_version_info():
     ''' Returns a triple of the current version (of the pip-package cryptoadvance.specter and 
         the latest version and whether you should upgrade 
     '''
     name="cryptoadvance.specter"
     try:
+        # fail right away if it's a binary
+        if getattr(sys, 'frozen', False):
+            raise RuntimeError("Using frozen binary, verision unavailable")
         latest_version = str(subprocess.run([sys.executable, '-m', 'pip', 'install', '{}==random'.format(name)], capture_output=True, text=True))
         latest_version = latest_version[latest_version.find('(from versions:')+15:]
         latest_version = latest_version[:latest_version.find(')')]
@@ -198,14 +217,17 @@ def get_version_info():
 
         current_version = str(subprocess.run([sys.executable, '-m', 'pip', 'show', '{}'.format(name)], capture_output=True, text=True))
         current_version = current_version[current_version.find('Version:')+8:]
-        current_version = current_version[:current_version.find('\\n')].replace(' ','') 
-
+        current_version = current_version[:current_version.find('\\n')].replace(' ','')
+        # master?
+        if not re.search(r"v([\d+]).([\d+]).([\d+]).*", current_version):
+            return current_version, latest_version, False
         return current_version, latest_version, latest_version != current_version
     except:
         # if pip is not installed or we are using python3.6 or below
         # we just don't show the version
         return "Unknown version", "Unknown version", False
 
+
 def get_users_json(specter):
     users = [
         {
@@ -226,11 +248,13 @@ def get_users_json(specter):
         save_users_json(specter, users)
     return users
 
+
 def save_users_json(specter, users):
     with fslock:
         with open(os.path.join(specter.data_folder, 'users.json'), "w") as f:
             f.write(json.dumps(users, indent=4))
 
+
 def hwi_get_config(specter):
     config = {
         'whitelisted_domains': 'http://127.0.0.1:25441/'
@@ -246,6 +270,7 @@ def hwi_get_config(specter):
         save_hwi_bridge_config(specter, config)
     return config
 
+
 def save_hwi_bridge_config(specter, config):
     if 'whitelisted_domains' in config:
         whitelisted_domains = ''
@@ -259,6 +284,7 @@ def save_hwi_bridge_config(specter, config):
         with open(os.path.join(specter.data_folder, 'hwi_bridge_config.json'), "w") as f:
             f.write(json.dumps(config, indent=4))
 
+
 def der_to_bytes(derivation):
     items = derivation.split("/")
     if len(items) == 0:
@@ -277,6 +303,7 @@ def der_to_bytes(derivation):
         res += index.to_bytes(4,'little')
     return res
 
+
 def get_devices_with_keys_by_type(app, cosigners, wallet_type):
     devices = []
     prefix = "tpub"
@@ -289,6 +316,7 @@ def get_devices_with_keys_by_type(app, cosigners, wallet_type):
         devices.append(device)
     return devices
 
+
 def sort_descriptor(cli, descriptor, index=None, change=False):
     descriptor = descriptor.replace("sortedmulti", "multi")
     if index is not None:
@@ -328,12 +356,14 @@ def sort_descriptor(cli, descriptor, index=None, change=False):
 
     return AddChecksum(desc)
 
+
 def hash_password(password):
     """Hash a password for storing."""
     salt = binascii.b2a_base64(hashlib.sha256(os.urandom(60)).digest()).strip()
     pwdhash = binascii.b2a_base64(hashlib.pbkdf2_hmac('sha256', password.encode('utf-8'), salt, 10000)).strip().decode()
     return { 'salt': salt.decode(), 'pwdhash': pwdhash }
 
+
 def verify_password(stored_password, provided_password):
     """Verify a stored password against one provided by user"""
     pwdhash = hashlib.pbkdf2_hmac('sha256', 
@@ -342,6 +372,7 @@ def verify_password(stored_password, provided_password):
                                   10000)
     return pwdhash == binascii.a2b_base64(stored_password['pwdhash'])
 
+
 def clean_psbt(b64psbt):
     psbt = PSBT()
     psbt.deserialize(b64psbt)
@@ -350,10 +381,12 @@ def clean_psbt(b64psbt):
             inp.non_witness_utxo = None
     return psbt.serialize()
 
+
 def bcur2base64(encoded):
     raw = bcur_decode(encoded.split("/")[-1])
     return binascii.b2a_base64(raw).strip()
 
+
 def get_txid(tx):
     b = BytesIO(bytes.fromhex(tx))
     t = CTransaction()
@@ -363,6 +396,21 @@ def get_txid(tx):
     t.rehash()
     return t.hash
 
+
+def get_startblock_by_chain(specter):
+    if specter.info['chain'] == "main":
+        if not specter.info['pruned'] or specter.info['pruneheight'] < 481824:
+            startblock = 481824
+        else:
+            startblock = specter.info['pruneheight']
+    else:
+        if not specter.info['pruned']:
+            startblock = 0
+        else:
+            startblock = specter.info['pruneheight']
+    return startblock
+
+
 # Hot wallet helpers
 def generate_mnemonic(strength=256):
         # Generate words list
@@ -370,43 +418,21 @@ def generate_mnemonic(strength=256):
         words = mnemo.generate(strength=strength)
         return words
 
-# From https://github.com/trezor/python-mnemonic/blob/ad06157e21fc2c2145c726efbfdcf69df1350061/mnemonic/mnemonic.py#L246
-# Refactored code segments from <https://github.com/keis/base58>
-def b58encode(v: bytes) -> str:
-    alphabet = "123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz"
 
-    p, acc = 1, 0
-    for c in reversed(v):
-        acc += p * c
-        p = p << 8
-
-    string = ""
-    while acc:
-        acc, idx = divmod(acc, 58)
-        string = alphabet[idx : idx + 1] + string
-    return string
-# We need to copy it like this because HWI uses it as a dependency, but requires v0.18 which doesn't have this function.
-def seed_to_hd_master_key(seed, testnet=False) -> str:
-    if len(seed) != 64:
-        raise ValueError("Provided seed should have length of 64")
-
-    # Compute HMAC-SHA512 of seed
-    seed = hmac.new(b"Bitcoin seed", seed, digestmod=hashlib.sha512).digest()
-
-    # Serialization format can be found at: https://github.com/bitcoin/bips/blob/master/bip-0032.mediawiki#Serialization_format
-    xprv = b"\x04\x88\xad\xe4"  # Version for private mainnet
-    if testnet:
-        xprv = b"\x04\x35\x83\x94"  # Version for private testnet
-    xprv += b"\x00" * 9  # Depth, parent fingerprint, and child number
-    xprv += seed[32:]  # Chain code
-    xprv += b"\x00" + seed[:32]  # Master key
-
-    # Double hash using SHA256
-    hashed_xprv = hashlib.sha256(xprv).digest()
-    hashed_xprv = hashlib.sha256(hashed_xprv).digest()
-
-    # Append 4 bytes of checksum
-    xprv += hashed_xprv[:4]
-
-    # Return base58
-    return b58encode(xprv)
+# Transaction processing helpers
+def parse_utxo(wallet, utxo):
+    for tx in utxo:
+        tx_data = wallet.cli.gettransaction(tx['txid'])
+        tx['time'] = tx_data['time']
+        if (len(tx_data['details']) > 1):
+            for details in tx_data['details']:
+                if details['category'] != 'send':
+                    tx['category'] = details['category']
+                    break
+        else:    
+            tx['category'] = tx_data['details'][0]['category']
+        if 'confirmations' in tx_data:
+            tx['confirmations'] = tx_data['confirmations']
+        else:
+            tx['confirmations'] = 0
+    return utxo
