@@ -18,8 +18,8 @@ def get_default_datadir():
         datadir = os.path.join(os.environ['HOME'], ".bitcoin")
     return datadir
 
-def get_rpcconfig():
-    path = get_default_datadir()
+
+def get_rpcconfig(datadir=get_default_datadir()):
     config = {
         "bitcoin.conf": {
             "default": {},
@@ -29,10 +29,10 @@ def get_rpcconfig():
         },
         "cookies": [],
     }
-    if not os.path.isdir(path): # we don't know where to search for files
+    if not os.path.isdir(datadir):  # we don't know where to search for files
         return config
     # load content from bitcoin.conf
-    bitcoin_conf_file = os.path.join(path, "bitcoin.conf")
+    bitcoin_conf_file = os.path.join(datadir, "bitcoin.conf")
     if os.path.exists(bitcoin_conf_file):
         try:
             with open(bitcoin_conf_file, 'r') as f:
@@ -48,16 +48,16 @@ def get_rpcconfig():
                         continue
                     k, v = line.split('=', 1)
                     current[k.strip()] = v.strip()
-        except:
+        except Exception:
             print("Can't open %s file" % bitcoin_conf_file)
     folders = {
-        "main": "", 
+        "main": "",
         "test": "testnet3",
         "regtest": "regtest",
         "signet": "signet",
     }
     for chain in folders:
-        fname = os.path.join(path, folders[chain], ".cookie")
+        fname = os.path.join(datadir, folders[chain], ".cookie")
         if os.path.exists(fname):
             try:
                 with open(fname, 'r') as f:
@@ -73,9 +73,10 @@ def get_rpcconfig():
                 print("Can't open %s file" % fname)
     return config
 
-def get_configs(config=None):
+
+def get_configs(config=None, datadir=get_default_datadir()):
     if config is None:
-        config = get_rpcconfig()
+        config = get_rpcconfig(datadir=datadir)
     confs = []
     default = {}
     for network in config["bitcoin.conf"]:
@@ -101,24 +102,25 @@ def get_configs(config=None):
         confs.append(o)
     return confs
 
-def detect_cli_confs(config=None):
+
+def detect_cli_confs(config=None, datadir=get_default_datadir()):
     if config is None:
-        config = get_rpcconfig()
+        config = get_rpcconfig(datadir=datadir)
     rpcconfs = get_configs(config)
     cli_arr = []
     for conf in rpcconfs:
         cli_arr.append(conf)
     return cli_arr
 
-def autodetect_cli_confs(port=None):
+
+def autodetect_cli_confs(datadir=get_default_datadir(), port=None):
     if port == "":
         port = None
     if port is not None:
         port = int(port)
-    conf_arr = detect_cli_confs()
+    conf_arr = detect_cli_confs(datadir=datadir)
     available_conf_arr = []
     if len(conf_arr) > 0:
-        print("trying %d different configs" % len(conf_arr))
         for conf in conf_arr:
             cli = BitcoinCLI(**conf)
             if port is not None:
@@ -131,9 +133,6 @@ def autodetect_cli_confs(port=None):
                 pass
             except Exception as e:
                 pass
-    else:
-        print("Bitcoin-cli not found :(")
-    print("Detected %d bitcoin daemons" % len(available_conf_arr))
     return available_conf_arr
 
 class RpcError(Exception):
@@ -148,10 +147,14 @@ class RpcError(Exception):
     '''
     def __init__(self, message, response):
         super(Exception, self).__init__(message)
-        self.status_code = response.status_code
         try:
-            error = json.loads(response.text)
             self.status_code = response.status_code
+            error = response.json()
+        except:
+            # ok already a dict
+            self.status_code = 500
+            error = response
+        try:
             self.error_code = error['error']['code']
             self.error_msg = error['error']['message']
         except Exception as e:
@@ -159,6 +162,7 @@ class RpcError(Exception):
 
 
 class BitcoinCLI:
+    counter = 0
     def __init__(self, user, passwd, host="127.0.0.1", port=8332, protocol="http", path="", timeout=None, **kwargs):
         path = path.replace("//","/") # just in case
         self.user = user
@@ -169,16 +173,16 @@ class BitcoinCLI:
         self.path = path
         self.timeout = timeout
         self.r = None
-        def wallet(name=""):
-            return BitcoinCLI(user=self.user,
-                          passwd=self.passwd,
-                          port=self.port,
-                          protocol=self.protocol,
-                          host=self.host,
-                          path="{}/wallet/{}".format(self.path, name),
-                          timeout=self.timeout
-            )
-        self.wallet = wallet
+
+    def wallet(self, name=""):
+        return BitcoinCLI(user=self.user,
+                      passwd=self.passwd,
+                      port=self.port,
+                      protocol=self.protocol,
+                      host=self.host,
+                      path="{}/wallet/{}".format(self.path, name),
+                      timeout=self.timeout
+        )
 
     @property
     def url(self):
@@ -193,41 +197,69 @@ class BitcoinCLI:
             return False
 
     def clone(self):
-        ''' returns a clone of self. Usefull if you want to mess with the properties '''
-        return BitcoinCLI(self.user, self.passwd, self.host, self.port, self.protocol, self.path, self.timeout)
+        """
+            Returns a clone of self.
+            Usefull if you want to mess with the properties
+        """
+        return BitcoinCLI(
+            self.user,
+            self.passwd,
+            self.host,
+            self.port,
+            self.protocol,
+            self.path,
+            self.timeout
+        )
+
+    def multi(self, calls: list, **kwargs):
+        """Makes batch request to Core"""
+        type(self).counter += len(calls)
+        # some debug info for optimizations
+        # methods = " ".join(list(dict.fromkeys([call[0] for call in calls])))
+        # wallet = self.path.split("/")[-1]
+        # print(f"{self.counter}: +{len(calls)} {wallet} {methods}")
+        headers = {'content-type': 'application/json'}
+        payload = [{
+            "method": method,
+            "params": args if args != [None] else [],
+            "jsonrpc": "2.0",
+            "id": i,
+        } for i, (method, *args) in enumerate(calls)]
+        timeout = self.timeout
+        if "timeout" in kwargs:
+            timeout = kwargs["timeout"]
+        url = self.url
+        if "wallet" in kwargs:
+            url = url+"/wallet/{}".format(kwargs["wallet"])
+        r = requests.post(
+            url, data=json.dumps(payload), headers=headers, timeout=timeout)
+        self.r = r
+        if r.status_code != 200:
+            raise RpcError(
+                "Server responded with error code %d: %s" % (
+                    r.status_code, r.text
+                ), r
+            )
+        r = r.json()
+        return r
 
     def __getattr__(self, method):
-        # if hasattr(self, method):
-            # self.
-        headers = {'content-type': 'application/json'}
         def fn(*args, **kwargs):
-            payload = {
-                "method": method,
-                "params": args,
-                "jsonrpc": "2.0",
-                "id": 0,
-            }
-            timeout = self.timeout
-            if "timeout" in kwargs:
-                timeout = kwargs["timeout"]
-            url = self.url
-            if "wallet" in kwargs:
-                url = url+"/wallet/{}".format(kwargs["wallet"])
-            r = requests.post(
-                url, data=json.dumps(payload), headers=headers, timeout=timeout)
-            self.r = r
-            if r.status_code != 200:
-                raise RpcError("Server responded with error code %d: %s" % (r.status_code, r.text), r)
-            r = r.json()
+            r = self.multi([(method,*args)], **kwargs)[0]
             if r["error"] is not None:
-                raise Exception(r["error"])
+                raise RpcError("Request error: %s" % r["error"]["message"], r)
             return r["result"]
         return fn
 
+
 if __name__ == '__main__':
 
-    cli = BitcoinCLI("bitcoinrpc", "foi3uf092ury97iufhjf30982hf928uew9jd209j", port=18443)
-    
+    cli = BitcoinCLI(
+        "bitcoinrpc",
+        "foi3uf092ury97iufhjf30982hf928uew9jd209j",
+        port=18443
+    )
+
     print(cli.url)
 
     print(cli.getmininginfo())
