@@ -3,7 +3,7 @@ from time import time
 from hwilib.descriptor import AddChecksum
 from .device import Device
 from .key import Key
-from .helpers import decode_base58, der_to_bytes, get_xpub_fingerprint, sort_descriptor, fslock
+from .helpers import decode_base58, der_to_bytes, get_xpub_fingerprint, sort_descriptor, fslock, parse_utxo
 from hwilib.serializations import PSBT, CTransaction
 from io import BytesIO
 from .specter_error import SpecterError
@@ -11,6 +11,7 @@ import threading
 
 # a gap of 20 addresses is what many wallets do
 WALLET_CHUNK = 20
+wallet_tx_batch = 100
 
 class Wallet():
     def __init__(
@@ -34,7 +35,8 @@ class Wallet():
         fullpath,
         device_manager,
         manager,
-        old_format_detected = False
+        old_format_detected=False,
+        last_block=None,
     ):
         self.name = name
         self.alias = alias
@@ -49,25 +51,72 @@ class Wallet():
         self.recv_descriptor = recv_descriptor
         self.change_descriptor = change_descriptor
         self.keys = keys
-        self.devices = [(device if isinstance(device, Device) else device_manager.get_by_alias(device)) for device in devices]
+        self.devices = [
+            (
+                device
+                if isinstance(device, Device)
+                else device_manager.get_by_alias(device)
+            )
+            for device in devices
+        ]
         if None in self.devices:
-            raise Exception('A device used by this wallet could not have been found!')
+            raise Exception(
+                'A device used by this wallet could not have been found!'
+            )
         self.sigs_required = sigs_required
         self.pending_psbts = pending_psbts
         self.fullpath = fullpath
         self.manager = manager
-        self.cli = self.manager.cli.wallet(os.path.join(self.manager.cli_path, self.alias))
+        self.cli = self.manager.cli.wallet(
+            os.path.join(self.manager.cli_path, self.alias)
+        )
+        self.last_block = last_block
 
         if address == '':
             self.getnewaddress()
         if change_address == '':
             self.getnewaddress(change=True)
 
-        self.cli.scan_addresses(self)
         self.getdata()
-        if old_format_detected:
+        self.update()
+        if old_format_detected or self.last_block != last_block:
             self.save_to_file()
 
+    def update(self):
+        self.get_balance()
+        self.check_addresses()
+        self.get_info()
+
+    def check_addresses(self):
+        """Checking the gap limit is still ok"""
+        if self.last_block is None:
+            obj = self.cli.listsinceblock()
+            txs = obj["transactions"]
+            last_block = obj["lastblock"]
+        else:
+            obj = self.cli.listsinceblock(self.last_block)
+            txs = obj["transactions"]
+            last_block = obj["lastblock"]
+        addresses = [tx["address"] for tx in txs]
+        # remove duplicates
+        addresses = list(dict.fromkeys(addresses))
+        if len(addresses) > 0:
+            # prepare rpc call
+            calls = [("getaddressinfo",addr) for addr in addresses]
+            # extract results
+            res = [r["result"] for r in self.cli.multi(calls)]
+            # extract last two indexes of hdkeypath
+            paths = [d["hdkeypath"].split("/")[-2:] for d in res if "hdkeypath" in d]
+            # get change and recv addresses
+            max_recv = max([int(p[1]) for p in paths if p[0]=="0"], default=-1)
+            max_change = max([int(p[1]) for p in paths if p[0]=="1"], default=-1)
+            # these calls will happen only if current addresses are used
+            while max_recv >= self.address_index:
+                self.getnewaddress(change=False)
+            while max_change >= self.change_index:
+                self.getnewaddress(change=True)
+        self.last_block = last_block
+
     @staticmethod
     def parse_old_format(wallet_dict, device_manager):
         old_format_detected = False
@@ -112,6 +161,7 @@ class Wallet():
         sigs_required = wallet_dict['sigs_required'] if 'sigs_required' in wallet_dict else 1
         pending_psbts = wallet_dict['pending_psbts'] if 'pending_psbts' in wallet_dict else {}
         fullpath = wallet_dict['fullpath'] if 'fullpath' in wallet_dict else default_fullpath
+        last_block = wallet_dict['last_block'] if 'last_block' in wallet_dict else None
 
         wallet_dict = Wallet.parse_old_format(wallet_dict, device_manager)
 
@@ -144,25 +194,36 @@ class Wallet():
             fullpath,
             device_manager,
             manager,
-            old_format_detected=wallet_dict['old_format_detected']
+            old_format_detected=wallet_dict['old_format_detected'],
+            last_block=last_block
         )
 
+    def get_info(self):
+        try:
+            self.info = self.cli.getwalletinfo()
+        except Exception:
+            self.info = {}
+
     def getdata(self):
         try:
-            self.utxo = self.cli.listunspent(0)
-        except:
+            self.utxo = parse_utxo(self, self.cli.listunspent(0))
+        except Exception:
             self.utxo = []
-        try:
-            self.transactions = self.cli.listtransactions("*", 1000, 0, True)
-        except:
-            self.transactions = []
-        try:
-            self.info = self.cli.getwalletinfo()
-        except:
-            self.info = {}
+        self.get_info()
         # TODO: Should do the same for the non change address (?)
         # check if address was used already
-        value_on_address = self.cli.getreceivedbyaddress(self.change_address, 0)
+        try:
+            value_on_address = self.cli.getreceivedbyaddress(
+                self.change_address,
+                0
+            )
+        except:
+            # Could happen if address not in wallet (wallet was imported)
+            # try adding keypool
+            self.keypoolrefill(0, end=self.keypool, change=False)
+            self.keypoolrefill(0, end=self.change_keypool, change=True)
+            value_on_address = 0
+
         # if not - just return
         if value_on_address > 0:
             self.change_index += 1
@@ -188,6 +249,8 @@ class Wallet():
             "sigs_required": self.sigs_required,
             "pending_psbts": self.pending_psbts,
             "fullpath": self.fullpath,
+            "last_block": self.last_block,
+            "blockheight": self.blockheight
         }
 
     def save_to_file(self):
@@ -238,18 +301,25 @@ class Wallet():
         self.cli.lockunspent(False, psbt["tx"]["vin"])
         self.save_to_file()
 
-    @property
-    def txlist(self):
-        ''' The last 1000 transactions for that wallet - filtering out change addresses transactions and duplicated transactions (except for self-transfers)
-            This list is used for the wallet `txs` tab to list the wallet transacions.
-        '''
-        txidlist = []
-        txlist = []
-        for tx in self.transactions:
-            if tx["is_change"] == False and (tx["is_self"] or tx["txid"] not in txidlist):
-                txidlist.append(tx["txid"])
-                txlist.append(tx)
-        return txlist
+    def txlist(self, idx, wallet_tx_batch=100):
+        try:
+            cli_txs = self.cli.listtransactions("*", wallet_tx_batch + 2, wallet_tx_batch * idx, True) # get batch + 2 to make sure you have information about send
+            cli_txs.reverse()
+            transactions = cli_txs[:wallet_tx_batch]
+        except:
+            return []
+        txids = []
+        result = []
+        for tx in transactions:
+            if 'confirmations' not in tx:
+                tx['confirmations'] = 0
+            if len([_tx for _tx in cli_txs if (_tx['txid'] == tx['txid'] and _tx['address'] == tx['address'])]) > 1:
+                continue # means the tx is duplicated (change), continue
+
+            txids.append(tx["txid"])
+            result.append(tx)
+
+        return result
 
     @property
     def rescan_progress(self):
@@ -263,9 +333,20 @@ class Wallet():
 
     @property
     def blockheight(self):
-        if len(self.transactions) > 0 and 'block_height' in self.transactions[0]:
-            return self.transactions[0]['block_height'] - 101 # To ensure coinbase transactions are indexed properly
-        return self.cli.getblockcount()
+        txs = self.cli.listtransactions("*", 100, 0, True)
+        i = 0
+        while (len(txs) == 100):
+            i += 1
+            next_txs = self.cli.listtransactions("*", 100, i * 100, True)
+            if (len(next_txs) > 0):
+                txs = next_txs
+            else:
+                break
+        current_blockheight = self.cli.getblockcount()
+        if len(txs) > 0 and 'confirmations' in txs[0]:
+            blockheight = current_blockheight - txs[0]['confirmations'] - 101 # To ensure coinbase transactions are indexed properly
+            return 0 if blockheight < 0 else blockheight # To ensure regtest don't have negative blockheight
+        return current_blockheight
 
     @property
     def account_map(self):
@@ -304,13 +385,12 @@ class Wallet():
             return addr
         return self.cli.deriveaddresses(desc, [index, index + 1])[0]
 
-    @property
-    def balance(self):
+    def get_balance(self):
         try:
-            balance = self.cli.getbalances()["watchonly"]
+            self.balance = self.cli.getbalances()["watchonly"]
         except:
-            balance = { "trusted": 0, "untrusted_pending": 0 }
-        return balance
+            self.balance = { "trusted": 0, "untrusted_pending": 0 }
+        return self.balance
 
     def keypoolrefill(self, start, end=None, change=False):
         if end is None:
@@ -350,7 +430,12 @@ class Wallet():
                 arg.pop("range")
                 batch = []
                 for i in range(start, end):
-                    sorted_desc = sort_descriptor(self.cli, desc, index=i, change=change)
+                    sorted_desc = sort_descriptor(
+                        self.cli,
+                        desc,
+                        index=i,
+                        change=change
+                    )
                     # create fresh object
                     obj = {}
                     obj.update(arg)
@@ -363,18 +448,18 @@ class Wallet():
             self.keypool = end
         self.save_to_file()
         return end
-    
-    def tx_on_address(self, address):
-        txlist = [tx for tx in self.transactions if tx["address"] == address]
-        return len(txlist)
+
+    def utxo_on_address(self, address):
+        utxo = [tx for tx in self.utxo if tx["address"] == address]
+        return len(utxo)
 
     def balance_on_address(self, address):
         balancelist = [utxo["amount"] for utxo in self.utxo if utxo["address"] == address]
         return sum(balancelist)
 
-    def tx_on_label(self, label):
-        txlist = [tx for tx in self.transactions if self.getlabel(tx["address"]) == label]
-        return len(txlist)
+    def utxo_on_label(self, label):
+        utxo = [tx for tx in self.utxo if self.getlabel(tx["address"]) == label]
+        return len(utxo)
 
     def balance_on_label(self, label):
         balancelist = [utxo["amount"] for utxo in self.utxo if self.getlabel(utxo["address"]) == label]
@@ -385,36 +470,17 @@ class Wallet():
             [address for address in (self.addresses + self.change_addresses) if self.getlabel(address) == label]
         ))
 
-    def is_tx_spent(self, txid):
-        return txid in [utxo["txid"] for utxo in self.utxo]
-
     @property
-    def tx_on_current_address(self):
-        return self.tx_on_address(self.address)
+    def is_current_address_used(self):
+        return self.balance_on_address(self.address) > 0
 
     @property
     def utxo_addresses(self):
-        return list(dict.fromkeys([
-            utxo["address"] for utxo in 
-            sorted(
-                self.utxo,
-                key = lambda utxo: next(
-                    tx for tx in self.transactions if tx["txid"] == utxo["txid"]
-                )["time"]
-            )
-        ]))
+        return list(dict.fromkeys([utxo["address"] for utxo in sorted(self.utxo, key = lambda utxo: utxo["time"])]))
 
     @property
     def utxo_labels(self):
-        return list(dict.fromkeys([
-            self.getlabel(utxo["address"]) for utxo in 
-            sorted(
-                self.utxo,
-                key = lambda utxo: next(
-                    tx for tx in self.transactions if tx["txid"] == utxo["txid"]
-                )["time"]
-            )
-        ]))
+        return list(dict.fromkeys([self.getlabel(utxo["address"]) for utxo in sorted(self.utxo, key = lambda utxo: utxo["time"])]))
 
     def setlabel(self, address, label):
         self.cli.setlabel(address, label)
@@ -422,7 +488,11 @@ class Wallet():
     def getlabel(self, address):
         address_info = self.cli.getaddressinfo(address)
         # Bitcoin Core version 0.20.0 has replaced the `label` field with `labels`, an array currently limited to a single item.
-        label = address_info["labels"][0] if "labels" in address_info and (isinstance(address_info["labels"], list) and len(address_info["labels"]) > 0) and "label" not in address_info else address
+        label = address_info["labels"][0] if (
+            "labels" in address_info 
+            and (isinstance(address_info["labels"], list) 
+                and len(address_info["labels"]) > 0) 
+            and "label" not in address_info) else address
         if label == "":
             label = address
         return address_info["label"] if "label" in address_info and address_info["label"] != "" else label
@@ -434,12 +504,15 @@ class Wallet():
 
     @property
     def fullbalance(self):
-        return self.balance["trusted"] + self.balance["untrusted_pending"]
+        balance = self.balance
+        return balance["trusted"] + balance["untrusted_pending"]
 
     @property
     def available_balance(self):
         locked_utxo = self.cli.listlockunspent()
-        balance = self.balance
+        # copy
+        balance = {}
+        balance.update(self.balance)
         for tx in locked_utxo:
             tx_data = self.cli.gettransaction(tx["txid"])
             raw_tx = self.cli.decoderawtransaction(tx_data["hex"])
@@ -466,31 +539,36 @@ class Wallet():
     def change_addresses(self):
         return [self.get_address(idx, change=True) for idx in range(0, self.change_index + 1)]
 
+    @property
+    def wallet_addresses(self):
+        return self.addresses + self.change_addresses
+
     @property
     def labels(self):
         return list(dict.fromkeys([self.getlabel(addr) for addr in self.active_addresses]))
 
-    def createpsbt(self, address:str, amount:float, subtract:bool=False, fee_rate:float=0.0, fee_unit="SAT_B", selected_coins=[]):
+    def createpsbt(self, addresses:[str], amounts:[float], subtract:bool=False, fee_rate:float=0.0, fee_unit="SAT_B", selected_coins=[]):
         """
             fee_rate: in sat/B or BTC/kB. Default (None) bitcoin core sets feeRate automatically.
         """
-        if self.full_available_balance < amount:
+
+        if self.full_available_balance < sum(amounts):
             raise SpecterError('The wallet does not have sufficient funds to make the transaction.')
 
         if fee_unit not in ["SAT_B", "BTC_KB"]:
             raise ValueError('Invalid bitcoin unit')
 
         extra_inputs = []
-        if self.available_balance["trusted"] < amount:
+        if self.available_balance["trusted"] < sum(amounts):
             txlist = self.cli.listunspent(0, 0)
-            b = amount - self.available_balance["trusted"]
+            b = sum(amounts) - self.available_balance["trusted"]
             for tx in txlist:
                 extra_inputs.append({"txid": tx["txid"], "vout": tx["vout"]})
                 b -= tx["amount"]
                 if b < 0:
                     break
         elif selected_coins != []:
-            still_needed = amount
+            still_needed = sum(amounts)
             for coin in selected_coins:
                 coin_txid = coin.split(",")[0]
                 coin_vout = int(coin.split(",")[1])
@@ -522,7 +600,7 @@ class Wallet():
         # don't reuse change addresses - use getrawchangeaddress instead
         r = self.cli.walletcreatefundedpsbt(
             extra_inputs,           # inputs
-            [{address: amount}],    # output
+            [{addresses[i]: amounts[i]} for i in range(len(addresses))],    # output
             0,                      # locktime
             options,                # options
             True                    # replaceable
@@ -531,8 +609,8 @@ class Wallet():
         b64psbt = r["psbt"]
         psbt = self.cli.decodepsbt(b64psbt)
         psbt['base64'] = b64psbt
-        psbt["amount"] = amount
-        psbt["address"] = address
+        psbt["amount"] = amounts
+        psbt["address"] = addresses
         psbt["time"] = time()
         psbt["sigs_count"] = 0
         self.save_pending_psbt(psbt)
